#-----------------------------------------------------------------------------
#
# Thomas Thomassen
# thomas[at]thomthom[dot]net
#
#-----------------------------------------------------------------------------

require 'sketchup.rb'
require 'TT_Lib/core.rb'

#-----------------------------------------------------------------------------

module TT_Lib
 module Geom2D
	
	# Graham's Scan
	# Computational Geometry Algorythms and Appliations (ISBN-10: 3540779736)
	# Chapter 1.1
	#
	# @param [Array<Geom::Point3d>] points
	# @return [Array<Geom::Point3d>]
	# @since 1.1.0
	def self.convex_hull(points)
		return nil if points.length < 2
		
		points = self.sort_points_by_x_y(points)
		
		l_upper = [ points[0], points[1] ]
		2.upto(points.length - 1) { |i|
			l_upper << points[i]
			while l_upper.length > 2 && !self.right_turn?(l_upper.last(3))
				l_upper.delete_at(-2)
			end
		}
		
		l_lower = [ points[-1], points[-2] ]
		(points.length - 3).downto(0) { |i|
			l_lower << points[i]
			while l_lower.length > 2 && !self.right_turn?(l_lower.last(3))
				l_lower.delete_at(-2)
			end
		}
		l_lower.delete_at(0)
		l_lower.delete_at(-1)
		
		return l_upper + l_lower
	end
	
	
	# @param [Array<Geom::Point3d>] points
	# @return [Array<Geom::Point3d>]
	# @since 1.1.0
	def self.sort_points_by_x(points)
		return points.sort { |a,b| a.x <=> b.x }
	end
	
	
	# Lexicographic order x, y
	#
	# @param [Array<Geom::Point3d>] points
	# @return [Array<Geom::Point3d>]
	# @since 1.1.0
	def self.sort_points_by_x_y(points)
		return points.sort { |a,b| (a.x==b.x) ? a.y <=> b.y : a.x <=> b.x }
	end
	
	
	# @param [Array<Geom::Point3d>] points array of 3 +Geom::Point3d+ objects.
	# @since 1.1.0 - Renamed from +.right_turn+ to +.right_turn?+ at version 1.2.0
	def self.right_turn?(points)
		p, q, r = points
		array = [ 1, p.x, p.y,
				  1, q.x, q.y,
				  1, r.x, r.y ]
		return self.determinant_3x3(array) < 0.0
	end
	
	# @param [Array<Number>] array
	# @return [Number]
	# @since 1.1.0
	def self.determinant_3x3(array)
		a,b,c,d,e,f,g,h,i = array
		return a*e*i - a*f*h + b*f*g - b*d*i + c*d*h - c*e*g
	end
	
 end # module Geom2D
end # module TT_Lib