#-----------------------------------------------------------------------------
#
# Thomas Thomassen
# thomas[at]thomthom[dot]net
#
#-----------------------------------------------------------------------------

require 'sketchup.rb'
require 'TT_Lib/core.rb'

#-----------------------------------------------------------------------------

module TT_Lib
 module Geom3D
	
	# Returns +plane+ in the format +[ point3d, vector3d ]+.
	#
	# @param [Array<Geom::Point3d, Geom::Vector3d>, Array<Number, Number, Number, Number>] plane
	#
	# @return [Array<Geom::Point3d, Geom::Vector3d>]
	#
	# @since 1.2.0
	def self.normalize_plane(plane)
		return plane if plane.length == 2
		a, b, c, d = plane
		v = Geom::Vector3d.new(a,b,c)
		p = ORIGIN.offset(v.reverse, d)
		return [p, v]
	end
	
	# Check if all points in the array are on the same plane.
	#
	# @param [Array<Geom::Point3d>] points
	#
	# @since 1.1.0
	def self.planar_points?(points)
		return false if points.length < 3
		v1 = points[0].vector_to(points[1])
		v2 = points[0].vector_to(points[2])
		return false unless v1.valid? && v2.valid?
		vector = v1 * v2
		plane = [ points[0], vector ]
		3.upto(points.length-1) { |i|
			return false unless points[i].on_plane?(plane)
		}
		return true
	end
  
  # (!) Document.
  def self.faces_coplanar?(face1, face2)
    vs = face1.vertices + face2.vertices
    plane = Geom.fit_plane_to_points(vs)
    vs.each { |v| 
      return false unless v.position.on_plane?(plane)
    }
    return true
  end
	
	# Creates a set of +Geom::Point3d+ objects for an arc.
	#
	# @param [Geom::Point3d] center
	# @param [Geom::Vector3d] xaxis
	# @param [Geom::Vector3d] normal
	# @param [Number] radius
	# @param [Float] start_angle in radians
	# @param [Float] start_angle in radians
	# @param [Integer] num_segments
	#
	# @return [Array<Geom::Point3d>]
	#
	# @since 1.1.0 - Fixed (1.2.0): added missing +self+.
	def self.arc(center, xaxis, normal, radius, start_angle, end_angle, num_segments = 12)
		# Generate the first point.
		t = Geom::Transformation.rotation(center, normal, start_angle )
		points = []
		points << center.offset(xaxis, radius).transform(t)
		# Prepare a transformation we can repeat on the last entry in point to complete the arc.
		t = Geom::Transformation.rotation(center, normal, (end_angle - start_angle) / num_segments )
		1.upto(num_segments) { |i|
			points << points.last.transform(t)
		}
		return points
	end
	
	# Creates a set of +Geom::Point3d+ objects for an circle.
	#
	# @param [Geom::Point3d] center
	# @param [Geom::Vector3d] normal
	# @param [Number] radius
	# @param [Integer] num_segments
	#
	# @return [Array<Geom::Point3d>]
	#
	# @since 1.1.0 - Fixed (1.2.0): added missing +self+.
	def self.circle(center, normal, radius, num_segments)
		points = arc(center, normal.axes.x, normal, radius, 0.0, Math::PI * 2, num_segments)
		points.pop
		return points
	end
	
	# Calculates the number of segments in an arc given the segments of a full circle. This 
	# will give a close visual quality of the arcs and circles.
	#
	# @param [Float] angle in radians
	# @param [Integer] full_circle_segments
	# @param [Boolean] force_even useful to ensure the segmented arc's
	#   apex hits the apex of the real arc
	#
	# @return [Integer]
	#
	# @since 1.1.0 - Fixed (1.2.0): added missing +self+.
	def self.arc_segments(angle, full_circle_segments, force_even = false)
		segments = (full_circle_segments * (angle / (Math::PI * 2))).to_i
		segments += 1 if force_even && segments % 2 > 0 # if odd
		return segments
	end

 end # module Geom3D
end # module TT_Lib