#-----------------------------------------------------------------------------
#
# Thomas Thomassen
# thomas[at]thomthom[dot]net
#
#-----------------------------------------------------------------------------

require 'sketchup.rb'

#-----------------------------------------------------------------------------


# UTF-8 subclass of +String+. Methods changed to be UTF-8 compatible is made to
# match the original method as closely as possible. But some have not been fully
# implemented.
#
# @todo Replace remaining methods that needs to be UTF-8 compatible.
# @todo Implement all argument variants.
class UTF8 < String

	# Rewire the String.length to be the UTF-8 string in bytes
	# @since 1.2.0
	alias :bytes :size
	
	# @return [Integer] the number of UTF-8 characters.
	# @since 1.2.0
	def length
		return self.unpack('U*').length
	end
	
	# @overload [](index)
	#   @param [Integer] index
	#   @return [UTF8]
	# @overload [](start, length)
	#   @param [Integer] start
	#   @param [Integer] length
	#   @return [UTF8]
	# @overload [](range)
	#   @param [Range] range
	#   @return [UTF8]
	#
	# @todo Implement all argument options that the native String object does.
	#
	# @return [UTF8]
	# @since 1.2.0
	def [](*args)
		if args.length == 1
			u = [ self.unpack('U*')[args.first] ]
		elsif args.length == 2
			u = self.unpack('U*')[args.first, args.last]
		else
			raise ArgumentError, 'Wrong number of arguments.'
		end
		return u.pack('U*')
	end
	alias :slice :[]
	
	# @overload [](index)
	#   @param [Integer] index
	#   @return [UTF8]
	# @overload [](start, length)
	#   @param [Integer] start
	#   @param [Integer] length
	#   @return [UTF8]
	# @overload [](range)
	#   @param [Range] range
	#   @return [UTF8]
	#
	# @todo Implement all argument options that the native String object does.
	#
	# @return [UTF8]
	# @since 1.2.0
	def []=(*args)
		u = self.unpack('U*')
		value = args.pop.unpack('U*')
		if args.length == 1
			u[args.first, 1] = value
		elsif args.length == 2
			u[args.first, args.last] = value
		else
			raise ArgumentError, 'Wrong number of arguments.'
		end
		self.replace( u.pack('U*') )
	end
	
	# @overload [](index)
	#   @param [Integer] index
	#   @return [UTF8]
	# @overload [](start, length)
	#   @param [Integer] start
	#   @param [Integer] length
	#   @return [UTF8]
	# @overload [](range)
	#   @param [Range] range
	#   @return [UTF8]
	#
	# @todo Implement all argument options that the native String object does.
	#
	# @return [UTF8]
	# @since 1.2.0
	def slice!(*args)
		if args.length == 1
			u = self.unpack('U*')[args.first]
		elsif args.length == 2
			u = self.unpack('U*')[args.first, args.last]
		else
			raise ArgumentError, 'Wrong number of arguments.'
		end
		self.replace( u.pack('U*') )
	end
	
	# @see http://ruby-doc.org/core/classes/String.html#M000815 String.center
	#
	# @param [Integer] integer
	# @param [String] padstr
	#
	# @return [UTF8]
	# @since 1.2.0
	def center(integer, padstr = ' ')
		return self if integer <= self.length
		
		padstrw = padstr.unpack('U*')
		
		left  = ((integer.to_f - self.length) / 2).floor
		right = ((integer.to_f - self.length) / 2).ceil
		
		left_str  = padstrw * (left  / padstrw.length).floor
		right_str = padstrw * (right / padstrw.length).floor
		
		left_str  = left_str[0, left].pack('U*')
		right_str = right_str[0, right].pack('U*')
		
		return UTF8.new( left_str + self + right_str )
	end
	
	# Returns a new String with the given record separator removed from the end of str
	# (if present). If $/ has not been changed from the default Ruby record separator, then
	# chomp also removes carriage return characters (that is it will remove \n, \r, and \r\n). 
	#def chomp(separator=$/)
		# (!)
	#end
	
	#def chomp!(separator=$/)
		# (!)
	#end
	
	# Returns a new String with the last character removed. If the string ends with \r\n, both
	# characters are removed. Applying chop to an empty string returns an empty string.
	# String#chomp is often a safer alternative, as it leaves the string unchanged if it doesn't
	# end in a record separator. 
	#
	# @return [UTF8]
	# @since 1.2.0
	def chop
		u = self.unpack('U*')
		if self[-2, 2] == "\r\n"
			u.delete_at(-2)
			return UTF8.new( u.pack('U*') )
		else
			u.delete_at(-1)
			return UTF8.new( u.pack('U*') )
		end
	end
	
	# @return [UTF8]
	# @since 1.2.0
	def chop!
		self.replace( self.chop )
	end
	
	# @yield [chr]
	# @yieldreturn [UTF8] Unicode character
	# @return [UTF8]
	# @since 1.2.0
	def each_char
		self.unpack('U*').each { |c| yield( UTF8.new( [c].pack('U*') ) ) }
		return self
	end
	
	# @see http://ruby-doc.org/core/classes/String.html#M000773 String.insert
	#
	# @param [Integer] index
	# @param [String] other_str
	#
	# @return [UTF8]
	# @since 1.2.0
	def insert(index, other_str)
		return UTF8.new( self.unpack('U*').insert( other_str.unpack('U*') ).pack('U*') )
	end
	
	# @see http://ruby-doc.org/core/classes/String.html#M000813 String.ljust
	#
	# @param [Integer] index
	# @param [String] padstr
	#
	# @return [UTF8]
	# @since 1.2.0
	def ljust(integer, padstr=' ')
		return self if integer <= self.length
		
		padstrw = padstr.unpack('U*')
		pad = (integer.to_f - self.length).to_f
		pad_str  = padstrw * (pad / padstrw.length).floor
		pad_str  = pad_str[0, pad].pack('U*')
		
		return UTF8.new( self + pad_str )
	end
	
	# @see http://ruby-doc.org/core/classes/String.html#M000804 String.reverse
	#
	# @return [UTF8]
	# @since 1.2.0
	def reverse
		return UTF8.new( self.unpack('U*').reverse.pack('U*') )
	end
	
	# @see http://ruby-doc.org/core/classes/String.html#M000805 String.reverse!
	#
	# @return [UTF8]
	# @since 1.2.0
	def reverse!
		self.replace( self.reverse )
	end
	
	# @see http://ruby-doc.org/core/classes/String.html#M000814 String.rjust
	#
	# @param [Integer] index
	# @param [String] padstr
	#
	# @return [UTF8]
	# @since 1.2.0
	def rjust(integer, padstr=' ')
		return self if integer <= self.length
		
		padstrw = padstr.unpack('U*')
		pad = (integer.to_f - self.length).to_f
		pad_str  = padstrw * (pad / padstrw.length).floor
		pad_str  = pad_str[0, pad].pack('U*')
		
		return UTF8.new( pad_str + self )
	end
	
end #class UTF8